import numpy as np
import pandas as pd
from scipy.stats import norm
import unittest

from context import grama as gr
from context import data

X = gr.Intention()

##==============================================================================
## select and drop test functions
##==============================================================================

#       0     1      2     3       4      5      6      7     8     9
#   carat    cut color clarity  depth  table  price     x     y     z
#    0.23  Ideal     E     SI2   61.5   55.0    326  3.95  3.98  2.43


class TestSelect(unittest.TestCase):
    def test_select(self):
        df = data.df_diamonds[["carat", "cut", "price"]]
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select("carat", "cut", "price"))
        )
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_select(0, 1, 6)))
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_select(0, 1, "price")))
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select([0, X.cut], X.price))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(X.carat, X["cut"], X.price))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(X[["carat", "cut", "price"]]))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(X[["carat", "cut"]], X.price))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(X.iloc[:, [0, 1, 6]]))
        )
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_select([X.loc[:, ["carat", "cut", "price"]]])
            )
        )

    def test_select_inversion(self):
        df = data.df_diamonds.iloc[:, 3:]
        d = data.df_diamonds >> gr.tf_select(~X.carat, ~X.cut, ~X.color)
        self.assertTrue(df.equals(d))

    def test_drop(self):
        df = data.df_diamonds.drop(["carat", "cut", "price"], axis=1)
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop("carat", "cut", "price"))
        )
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop(0, 1, 6)))
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop(0, 1, "price")))
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop([0, X.cut], X.price)))
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(X.carat, X["cut"], X.price))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(X[["carat", "cut", "price"]]))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(X[["carat", "cut"]], X.price))
        )
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop(X.iloc[:, [0, 1, 6]])))
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_drop([X.loc[:, ["carat", "cut", "price"]]])
            )
        )

    def test_select_containing(self):
        df = data.df_diamonds[["carat", "cut", "color", "clarity", "price"]]
        assert df.equals(data.df_diamonds >> gr.tf_select(gr.contains("c")))

    def test_drop_containing(self):
        df = data.df_diamonds[["depth", "table", "x", "y", "z"]]
        assert df.equals(data.df_diamonds >> gr.tf_drop(gr.contains("c")))

    def test_select_matches(self):
        df = data.df_diamonds[["carat", "cut", "color", "clarity", "price"]]
        assert df.equals(data.df_diamonds >> gr.tf_select(gr.matches("^c[auol]|pri")))

    def test_drop_matches(self):
        df = data.df_diamonds[["depth", "table", "x", "y", "z"]]
        assert df.equals(data.df_diamonds >> gr.tf_drop(gr.matches("^c[auol]|p.i")))

    def test_select_startswith(self):
        df = data.df_diamonds[["carat", "cut", "color", "clarity"]]
        assert df.equals(data.df_diamonds >> gr.tf_select(gr.starts_with("c")))

    def test_drop_startswith(self):
        df = data.df_diamonds[["depth", "table", "price", "x", "y", "z"]]
        assert df.equals(data.df_diamonds >> gr.tf_drop(gr.starts_with("c")))

    def test_select_endswith(self):
        df = data.df_diamonds[["table", "price"]]
        assert df.equals(data.df_diamonds >> gr.tf_select(gr.ends_with("e")))

    def test_drop_endswith(self):
        df = data.df_diamonds.drop("z", axis=1)
        assert df.equals(data.df_diamonds >> gr.tf_drop(gr.ends_with("z")))

    def test_select_between(self):
        df = data.df_diamonds[["cut", "color", "clarity"]]
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_select(gr.columns_between(X.cut, X.clarity))
            )
        )
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_select(gr.columns_between("cut", "clarity"))
            )
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(gr.columns_between(1, 3)))
        )

        df = data.df_diamonds[["x", "y", "z"]]
        assert df.equals(data.df_diamonds >> gr.tf_select(gr.columns_between("x", 20)))

    def test_drop_between(self):
        df = data.df_diamonds[["carat", "z"]]
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_between("cut", "y")))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_between(X.cut, 8)))
        )

        df = data.df_diamonds[["carat", "cut"]]
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_between(X.color, 20)))
        )

    def test_select_from(self):
        df = data.df_diamonds[["x", "y", "z"]]
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(gr.columns_from("x")))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(gr.columns_from(X.x)))
        )
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_select(gr.columns_from(7))))
        self.assertTrue(
            data.df_diamonds[[]].equals(
                data.df_diamonds >> gr.tf_select(gr.columns_from(100))
            )
        )

    def test_drop_from(self):
        df = data.df_diamonds[["carat", "cut"]]
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_from("color")))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_from(X.color)))
        )
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_from(2))))
        self.assertTrue(
            data.df_diamonds[[]].equals(
                data.df_diamonds >> gr.tf_drop(gr.columns_from(0))
            )
        )

    def test_select_to(self):
        df = data.df_diamonds[["carat", "cut"]]
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(gr.columns_to("color")))
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_select(gr.columns_to(X.color)))
        )
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_select(gr.columns_to(2))))

    def test_drop_to(self):
        df = data.df_diamonds[["x", "y", "z"]]
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_to("x"))))
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_to(X.x))))
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_to(7))))

    def select_through(self):
        df = data.df_diamonds[["carat", "cut", "color"]]
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_select(gr.columns_to("color", inclusive=True))
            )
        )
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_select(gr.columns_to(X.color, inclusive=True))
            )
        )
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_select(gr.columns_to(2, inclusive=True))
            )
        )

    def drop_through(self):
        df = data.df_diamonds[["y", "z"]]
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_drop(gr.columns_to("x", inclusive=True))
            )
        )
        self.assertTrue(
            df.equals(
                data.df_diamonds >> gr.tf_drop(gr.columns_to(X.x, inclusive=True))
            )
        )
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_drop(gr.columns_to(7, inclusive=True)))
        )

    def test_select_if(self):
        # test 1: manually build data.df_diamonds subset where columns are numeric and
        # mean is greater than 3
        cols = list()
        for col in data.df_diamonds:
            try:
                if mean(data.df_diamonds[col]) > 3:
                    cols.append(col)
            except:
                pass
        df_if = data.df_diamonds[cols]
        self.assertTrue(
            df_if.equals(data.df_diamonds >> gr.tf_select_if(lambda col: mean(col) > 3))
        )
        # test 2: use and
        cols = list()
        for col in data.df_diamonds:
            try:
                if mean(data.df_diamonds[col]) > 3 and max(data.df_diamonds[col]) < 50:
                    cols.append(col)
            except:
                pass
        df_if = data.df_diamonds[cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_select_if(lambda col: mean(col) > 3 and max(col) < 50)
            )
        )
        # test 3: use or
        cols = list()
        for col in data.df_diamonds:
            try:
                if mean(data.df_diamonds[col]) > 3 or max(data.df_diamonds[col]) < 6:
                    cols.append(col)
            except:
                pass
        df_if = data.df_diamonds[cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_select_if(lambda col: mean(col) > 3 or max(col) < 6)
            )
        )
        # test 4: string operations - contain a specific string
        cols = list()
        for col in data.df_diamonds:
            try:
                if any(data.df_diamonds[col].str.contains("Ideal")):
                    cols.append(col)
            except:
                pass
        df_if = data.df_diamonds[cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_select_if(lambda col: any(col.str.contains("Ideal")))
            )
        )
        # test 5: get any text columns
        # uses the special '.' regex symbol to find any text value
        cols = list()
        for col in data.df_diamonds:
            try:
                if any(data.df_diamonds[col].str.contains(".")):
                    cols.append(col)
            except:
                pass
        df_if = data.df_diamonds[cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_select_if(lambda col: any(col.str.contains(".")))
            )
        )
        # test 6: is_numeric helper
        self.assertEqual(
            {"carat", "depth", "table", "price", "x", "y", "z"},
            set(
                (
                    data.df_diamonds
                    >> gr.tf_select_if(gr.is_numeric)
                ).columns
            )
        )

    def test_drop_if(self):
        # test 1: returns a dataframe where any column does not have a mean greater than 3
        # this means numeric columns with mean less than 3, and also any non-numeric column
        # (since it does not have a mean)
        cols = list()
        for col in data.df_diamonds:
            try:
                if mean(data.df_diamonds[col]) > 3:
                    cols.append(col)
            except:
                pass
        inverse_cols = [col for col in data.df_diamonds if col not in cols]
        df_if = data.df_diamonds[inverse_cols]
        self.assertTrue(
            df_if.equals(data.df_diamonds >> gr.tf_drop_if(lambda col: mean(col) > 3))
        )
        # test 2: use and
        # return colums where both conditions are false:
        # the mean greater than 3, and max < 50
        # again, this will include non-numeric columns
        cols = list()
        for col in data.df_diamonds:
            try:
                if mean(data.df_diamonds[col]) > 3 and max(data.df_diamonds[col]) < 50:
                    cols.append(col)
            except:
                pass
        inverse_cols = [col for col in data.df_diamonds if col not in cols]
        df_if = data.df_diamonds[inverse_cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_drop_if(lambda col: mean(col) > 3 and max(col) < 50)
            )
        )
        # test 3: use or
        # this will return a dataframe where either of the two conditions are false:
        # the mean is greater than 3, or the max < 6
        cols = list()
        for col in data.df_diamonds:
            try:
                if mean(data.df_diamonds[col]) > 3 or max(data.df_diamonds[col]) < 6:
                    cols.append(col)
            except:
                pass
        inverse_cols = [col for col in data.df_diamonds if col not in cols]
        df_if = data.df_diamonds[inverse_cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_drop_if(lambda col: mean(col) > 3 or max(col) < 6)
            )
        )
        # test 4: string operations - contain a specific string
        # this will drop any columns if they contain the word 'Ideal'
        cols = list()
        for col in data.df_diamonds:
            try:
                if any(data.df_diamonds[col].str.contains("Ideal")):
                    cols.append(col)
            except:
                pass
        inverse_cols = [col for col in data.df_diamonds if col not in cols]
        df_if = data.df_diamonds[inverse_cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_drop_if(lambda col: any(col.str.contains("Ideal")))
            )
        )
        # test 5: drop any text columns
        # uses the special '.' regex symbol to find any text value
        cols = list()
        for col in data.df_diamonds:
            try:
                if any(data.df_diamonds[col].str.contains(".")):
                    cols.append(col)
            except:
                pass
        inverse_cols = [col for col in data.df_diamonds if col not in cols]
        df_if = data.df_diamonds[inverse_cols]
        self.assertTrue(
            df_if.equals(
                data.df_diamonds
                >> gr.tf_drop_if(lambda col: any(col.str.contains(".")))
            )
        )



from copy import deepcopy

TOL = 1e-5

DEFAULT_CVXPY_KWARGS = {
	'solver': 'CVXOPT',
	'reltol': 5e-10,
	'abstol' : 5e-10,
	'verbose': False,
	'kktsolver': 'robust', 
	'warm_start': True,
}

class Domain(object):
	r""" Abstract base class for arbitary domain shapes
	"""
	
	def copy(self):
		return deepcopy(self)
	
	################################################################################
	# Note that the following properities of the domain really should be determined 
	# by the class name, except for the one annoying exception: the TensorProductDomain
	# which is a child of Domain. These functions are here to provide access to this
	# information independent of the domain name.
	################################################################################

	@property
	def is_linquad_domain(self):
		r""" Returns true if the domain is purely specified by linear equality/inequality and convex quadratic constraints
		"""
		return self._is_linquad_domain()

	# By default, we assume the domain is not in this class unless the child class explicitly corrects this
	def _is_linquad_domain(self):
		return False

	@property
	def is_linineq_domain(self):
		r""" Returns True if the domain is specified by linear equality/inqueality constraints
		"""
		return self._is_linineq_domain()
	
	def _is_linineq_domain(self):
		return False

	@property
	def is_box_domain(self):
		r""" Returns True if the domain is specified only by bound constraints on each variable
		"""
		return self._is_box_domain()

	def _is_box_domain(self):
		return False



__all__ = [
    "add_pipe",
    "copy_meta",
    "custom_formatwarning",
    "lookup",
    "find_files",
    "hide_traceback",
    "param_dist",
    "pipe",
    "valid_dist",
    "tran_outer",
    "tf_outer",
]

import warnings
import sys
from functools import wraps
from inspect import signature
from numbers import Integral
from glob import glob
from numpy import empty
from os.path import join as pathjoin
from pandas import DataFrame, concat
from pandas.core.dtypes.common import is_object_dtype
from pandas._libs import (
    algos as libalgos,
    lib,
    properties,
)
from pandas.testing import assert_frame_equal
from scipy.stats import alpha, anglit, arcsine, argus, beta, betaprime, \
    bradford, burr, burr12, cauchy, chi, chi2, cosine, crystalball, dgamma, \
    dweibull, erlang, expon, exponnorm, exponweib, exponpow, f, fatiguelife, \
    fisk, foldcauchy, foldnorm, gaussian_kde, genlogistic, gennorm, genpareto, \
    genexpon, genextreme, gausshyper, gamma, gengamma, genhalflogistic, \
    gibrat, gompertz, gumbel_r, gumbel_l, halfcauchy, halflogistic, \
    halfnorm, halfgennorm, hypsecant, invgamma, invgauss, invweibull, \
    johnsonsb, johnsonsu, kappa4, kappa3, ksone, kstwobign, laplace, levy, \
    levy_l, levy_stable, logistic, loggamma, loglaplace, lognorm, lomax, \
    maxwell, mielke, moyal, nakagami, ncx2, ncf, nct, norm, norminvgauss, \
    pareto, pearson3, powerlaw, powerlognorm, powernorm, rdist, rayleigh, \
    rice, recipinvgauss, skewnorm, t, trapz, triang, truncexpon, truncnorm, \
    tukeylambda, uniform, vonmises, vonmises_line, wald, weibull_min, \
    weibull_max, wrapcauchy
from toolz import curry


## Scipy metadata
valid_dist = {
    "alpha": alpha,
    "anglit": anglit,
    "arcsine": arcsine,
    "argus": argus,
    "beta": beta,
    "betaprime": betaprime,
    "bradford": bradford,
    "burr": burr,
    "burr12": burr12,
    "cauchy": cauchy,
    "chi": chi,
    "chi2": chi2,
    "cosine": cosine,
    "crystalball": crystalball,
    "dgamma": dgamma,
    "dweibull": dweibull,
    "erlang": erlang,
    "expon": expon,
    "exponnorm": exponnorm,
    "exponweib": exponweib,
    "exponpow": exponpow,
    "f": f,
    "fatiguelife": fatiguelife,
    "fisk": fisk,
    "foldcauchy": foldcauchy,
    "foldnorm": foldnorm,
    # "frechet_r": frechet_r,
    # "frechet_l": frechet_l,
    "genlogistic": genlogistic,
    "gennorm": gennorm,
    "genpareto": genpareto,
    "genexpon": genexpon,
    "genextreme": genextreme,
    "gausshyper": gausshyper,
    "gamma": gamma,
    "gengamma": gengamma,
    "genhalflogistic": genhalflogistic,
    # "geninvgauss": geninvgauss,
    "gibrat": gibrat,
    "gompertz": gompertz,
    "gumbel_r": gumbel_r,
    "gumbel_l": gumbel_l,
    "halfcauchy": halfcauchy,
    "halflogistic": halflogistic,
    "halfnorm": halfnorm,
    "halfgennorm": halfgennorm,
    "hypsecant": hypsecant,
    "invgamma": invgamma,
    "invgauss": invgauss,
    "invweibull": invweibull,
    "johnsonsb": johnsonsb,
    "johnsonsu": johnsonsu,
    "kappa4": kappa4,
    "kappa3": kappa3,
    "ksone": ksone,
    "kstwobign": kstwobign,
    "laplace": laplace,
    "levy": levy,
    "levy_l": levy_l,
    "levy_stable": levy_stable,
    "logistic": logistic,
    "loggamma": loggamma,
    "loglaplace": loglaplace,
    "lognorm": lognorm,
    # "loguniform": loguniform,
    "lomax": lomax,
    "maxwell": maxwell,
    "mielke": mielke,
    "moyal": moyal,
    "nakagami": nakagami,
    "ncx2": ncx2,
    "ncf": ncf,
    "nct": nct,
    "norm": norm,
    "norminvgauss": norminvgauss,
    "pareto": pareto,
    "pearson3": pearson3,
    "powerlaw": powerlaw,
    "powerlognorm": powerlognorm,
    "powernorm": powernorm,
    "rdist": rdist,
    "rayleigh": rayleigh,
    "rice": rice,
    "recipinvgauss": recipinvgauss,
    "skewnorm": skewnorm,
    "t": t,
    "trapz": trapz,
    "triang": triang,
    "truncexpon": truncexpon,
    "truncnorm": truncnorm,
    "tukeylambda": tukeylambda,
    "uniform": uniform,
    "vonmises": vonmises,
    "vonmises_line": vonmises_line,
    "wald": wald,
    "weibull_min": weibull_min,
    "weibull_max": weibull_max,
    "wrapcauchy": wrapcauchy,
}

param_dist = {
    "alpha": ["a", "loc", "scale"],
    "anglit": ["loc", "scale"],
    "arcsine": ["loc", "scale"],
    "argus": ["chi", "loc", "scale"],
    "beta": ["a", "b", "loc", "scale"],
    "betaprime": ["a", "b", "loc", "scale"],
    "bradford": ["c", "loc", "scale"],
    "burr": ["c", "d", "loc", "scale"],
    "burr12": ["c", "d", "loc", "scale"],
    "cauchy": ["loc", "scale"],
    "chi": ["df", "loc", "scale"],
    "chi2": ["df", "loc", "scale"],
    "cosine": ["loc", "scale"],
    "crystalball": ["beta", "m", "loc", "scale"],
    "dgamma": ["a", "loc", "scale"],
    "dweibull": ["c", "loc", "scale"],
    "erlang": ["loc", "scale"],
    "expon": ["loc", "scale"],
    "exponnorm": ["K", "loc", "scale"],
    "exponweib": ["a", "c", "loc", "scale"],
    "exponpow": ["b", "loc", "scale"],
    "f": ["dfn", "dfd", "loc", "scale"],
    "fatiguelife": ["c", "loc", "scale"],
    "fisk": ["c", "loc", "scale"],
    "foldcauchy": ["c", "loc", "scale"],
    "foldnorm": ["c", "loc", "scale"],
    "frechet_r": ["c", "loc", "scale"],
    "frechet_l": ["c", "loc", "scale"],
    "genlogistic": ["c", "loc", "scale"],
    "gennorm": ["beta", "loc", "scale"],
    "genpareto": ["c", "loc", "scale"],
    "genexpon": ["a", "b", "c", "loc", "scale"],
    "genextreme": ["c", "loc", "scale"],
    "gausshyper": ["a", "b", "c", "z", "loc", "scale"],
    "gamma": ["a", "loc", "scale"],
    "gengamma": ["a", "c", "loc", "scale"],
    "genhalflogistic": ["c", "loc", "scale"],
    "geninvgauss": ["p", "b", "loc", "scale"],
    "gibrat": ["loc", "scale"],
    "gompertz": ["c", "loc", "scale"],
    "gumbel_r": ["loc", "scale"],
    "gumbel_l": ["loc", "scale"],
    "halfcauchy": ["loc", "scale"],
    "halflogistic": ["loc", "scale"],
    "halfnorm": ["loc", "scale"],
    "halfgennorm": ["beta", "loc", "scale"],
    "hypsecant": ["loc", "scale"],
    "invgamma": ["a", "loc", "scale"],
    "invgauss": ["mu", "loc", "scale"],
    "invweibull": ["c", "loc", "scale"],
    "johnsonsb": ["a", "b", "loc", "scale"],
    "johnsonsu": ["a", "b", "loc", "scale"],
    "kappa4": ["h", "k", "loc", "scale"],
    "kappa3": ["a", "loc", "scale"],
    "ksone": ["n", "loc", "scale"],
    "kstwobign": ["loc", "scale"],
    "laplace": ["loc", "scale"],
    "levy": ["loc", "scale"],
    "levy_l": ["loc", "scale"],
    "levy_stable": ["alpha", "beta", "loc", "scale"],
    "logistic": ["loc", "scale"],
    "loggamma": ["c", "loc", "scale"],
    "loglaplace": ["c", "loc", "scale"],
    "lognorm": ["s", "loc", "scale"],
    "loguniform": ["a", "b", "loc", "scale"],
    "lomax": ["c", "loc", "scale"],
    "maxwell": ["loc", "scale"],
    "mielke": ["k", "s", "loc", "scale"],
    "moyal": ["loc", "scale"],
    "nakagami": ["nu", "loc", "scale"],
    "ncx2": ["df", "nc", "loc", "scale"],
    "ncf": ["dfn", "dfd", "nc", "loc", "scale"],
    "nct": ["df", "nc", "loc", "scale"],
    "norm": ["loc", "scale"],
    "norminvgauss": ["a", "b", "loc", "scale"],
    "pareto": ["b", "loc", "scale"],
    "pearson3": ["skew", "loc", "scale"],
    "powerlaw": ["a", "loc", "scale"],
    "powerlognorm": ["c", "s", "loc", "scale"],
    "powernorm": ["c", "loc", "scale"],
    "rdist": ["c", "loc", "scale"],
    "rayleigh": ["loc", "scale"],
    "rice": ["b", "loc", "scale"],
    "recipinvgauss": ["mu", "loc", "scale"],
    "semicircular": ["loc", "scale"],
    "skewnorm": ["a", "loc", "scale"],
    "t": ["df", "loc", "scale"],
    "trapz": ["c", "d", "loc", "scale"],
    "triang": ["c", "loc", "scale"],
    "truncexpon": ["b", "loc", "scale"],
    "truncnorm": ["a", "b", "loc", "scale"],
    "tukeylambda": ["lam", "loc", "scale"],
    "uniform": ["loc", "scale"],
    "vonmises": ["kappa", "loc", "scale"],
    "vonmises_line": ["kappa", "loc", "scale"],
    "wald": ["loc", "scale"],
    "weibull_min": ["c", "loc", "scale"],
    "weibull_max": ["c", "loc", "scale"],
    "wrapcauchy": ["c", "loc", "scale"],
}


## Core helper functions
##################################################
## Metadata copy
def copy_meta(df_source, df_target):
    """Internal metadata copy tool

    Args:
        df_source (DataFrame): Original dataframe
        df_target (DataFrame): Target dataframe; receives metadata

    Returns:
        DataFrame: df_target with copied metadata
    """
    df_target._grouped_by = getattr(df_source, "_grouped_by", None)
    df_target._plot_info = getattr(df_source, "_plot_info", None)
    df_target._meta = getattr(df_source, "_meta", None)

    return df_target


## Pipe decorator
class pipe(object):
    __name__ = "pipe"

    def __init__(self, function):
        self.function = function
        # self.__doc__ = function.__doc__
        self.chained_pipes = []

    def __rshift__(self, other):
        assert isinstance(other, pipe)
        self.chained_pipes.append(other)
        return self

    def __rrshift__(self, other):
        other_copy = other.copy()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if isinstance(other, DataFrame):
                other_copy = copy_meta(other, other_copy)

        result = self.function(other_copy)

        for p in self.chained_pipes:
            result = p.__rrshift__(result)
        return result

    def __repr__(self):
        return (
            self.__name__
            + ": "
            + self.function.__name__
            + "\n Did you mean to use a full-prefix verb?"
        )

    def __call__(self, *args, **kwargs):
        return pipe(lambda x: self.function(x, *args, **kwargs))


## Pipe applicator
def add_pipe(fun):
    class NewPipe(pipe):
        __name__ = fun.__name__
        __doc__ = (
            fun.__doc__
            #"Pipe-enabled version of {}\n".format(fun)
            #+ "Inherited Signature: {}\n".format(signature(fun))
            #+ fun.__doc__
        )

    return NewPipe(fun)


## Monkey-patched warning fcn
def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return "Warning: " + str(msg) + "\n"


## DataFrame outer product
# --------------------------------------------------
@curry
def tran_outer(df, df_outer):
    r"""Outer merge

    Perform an outer-merge on two dataframes.

    Args:
        df (DataFrame): Data to merge
        df_outer (DataFrame): Data to merge; outer

    Returns:
        DataFrame: Merged data

    Examples::

        import grama as gr
        import pandas as pd
        df = pd.DataFrame(dict(x=[1,2]))
        df_outer = pd.DataFrame(dict(y=[3,4]))
        df_res = gr.tran_outer(df, df_outer)
        df_res
             x  y
          0  1  3
          1  2  3
          2  1  4
          3  2  4

    """
    # Check invariants
    if (df.shape[0] == 0) and (df_outer.shape[0] == 0):
        raise ValueError("At least one of df and df_outer must be non-empty")
    # Handle single-empty cases
    if (df.shape[0] == 0):
        return df_outer
    if (df_outer.shape[0] == 0):
        return df

    n_rows = df.shape[0]
    list_df = []

    for ind in range(df_outer.shape[0]):
        df_rep = concat([df_outer.iloc[[ind]]] * n_rows, ignore_index=True)
        list_df.append(concat((df, df_rep), axis=1))

    return concat(list_df, ignore_index=True)


tf_outer = add_pipe(tran_outer)


def lookup(df, row_labels, col_labels):
    r"""2D lookup function for a dataframe
    (Old Pandas Lookup Method)

    Args:
        df (DataFrame): DataFrame for lookup
        row_labels (List): Row labels to use for lookup.
        col_labels (List): Column labels to use for lookup.

    Returns:
        numpy.ndarray: Found values
    """
    n = len(row_labels)
    if n != len(col_labels):
        raise ValueError("Row labels must have same size as column labels")
    if not (df.index.is_unique and df.columns.is_unique):
        raise ValueError("DataFrame.lookup requires unique index and columns")

    thresh = 1000
    if not df._is_mixed_type or n > thresh:
        values = df.values
        ridx = df.index.get_indexer(row_labels)
        cidx = df.columns.get_indexer(col_labels)
        if (ridx == -1).any():
            raise KeyError("One or more row labels was not found")
        if (cidx == -1).any():
            raise KeyError("One or more column labels was not found")
        flat_index = ridx * len(df.columns) + cidx
        result = values.flat[flat_index]
    else:
        result = empty(n, dtype="O")
        for i, (r, c) in enumerate(zip(row_labels, col_labels)):
            print(r,c)
            result[i] = df._get_value(r, c)

    if is_object_dtype(result):
        result = lib.maybe_convert_objects(result)

    return result


# Suppress traceback
def hide_traceback():
    r"""Configure Jupyter to hide error traceback
    """
    ipython = get_ipython()

    def _hide_traceback(exc_tuple=None, filename=None, tb_offset=None,
                        exception_only=False, running_compiled_code=False):
        etype, value, tb = sys.exc_info()
        value.__cause__ = None  # suppress chained exceptions
        return ipython._showtraceback(etype, value, ipython.InteractiveTB.get_exception_only(etype, value))

    ipython.showtraceback = _hide_traceback

# Find files
def find_files(dir, ext, recursive=False):
    r"""Find files in a given directory

    Args:
        dir (str or list[str]): Path as string or list of directories (to be joined)
        ext (str): File extension

    Kwargs:
        recursive (bool): Search recursively?
    """
    ## Parse arguments
    # Handle dir provided as list
    if isinstance(dir, list):
        dir = pathjoin(*dir)
    # Ensure leading dot
    if ext[0] != ".":
        ext = "." + ext
    ## Find the files
    if recursive:
        files = glob(pathjoin(dir, "**", "*" + ext), recursive=True)
    else:
        files = glob(pathjoin(dir, "*" + ext), recursive=False)

    return files



import unittest

from context import grama as gr
from context import data
from test_evals import TestEvalInvariants
import numpy as np
from numpy import array, pi
from grama.models import make_pareto_random
from grama.fit import ft_gp

class TestEvalPND(unittest.TestCase):
    """Test implementation of eval_pnd
    """
    def setUp(self):
        ## Invariant test class
        self.inv_test = TestEvalInvariants()

    def test_eval_pnd(self):
        """ Test basic evaluation of pnd function
        """
        # Invariant checks
        self.inv_test.md_arg(
            gr.eval_pnd, 
            df_arg="df_train",
            df_test = self.inv_test.df,
            signs = {"y1":1, "y2":1},
            seed = 101)
        self.inv_test.df_arg_2(
            gr.eval_pnd, 
            df_args=["df_train", "df_test"],
            signs = {"y1":1, "y2":1},
            seed = 101)

        # Model to make Dataset
        md_true = make_pareto_random()
        # Create dataframe
        df_data = (
            md_true
            >> gr.ev_sample(n=2e3, seed=101, df_det="nom")
        )
        ## Select training set
        df_train = (
            df_data
            >> gr.tf_sample(n=10)
        )
        ## select test set
        df_test = (
            df_data
                >> gr.tf_anti_join(
                    df_train,
                    by=["x1", "x2"],
                )
                >> gr.tf_sample(n=200)
        )

        # Create fitted model
        md_fit = (
            df_train
            >> ft_gp(
                var=["x1", "x2"],
                out=["y1", "y2"],
            )
        )

        # Call eval_pnd
        df_pnd = (
            md_fit
            >> gr.ev_pnd(
                df_train,
                df_test,
                signs = {"y1":1, "y2":1},
                seed = 101
            )
        )

        # Test for correctness by shape
        self.assertTrue(len(df_pnd) == df_test.shape[0])
        # Test for correctness by # of outputs
        self.assertTrue(len(df_pnd.columns.values) == len(df_test.columns.values) + 2)

    def test_eval_3D(self):
        """ Test 3D case for eval_pnd()
        """
        # Model to make Dataset
        md_true = make_pareto_random(twoDim=False)

        # Create dataframe
        df_data = (
            md_true
            >> gr.ev_sample(n=2e3, seed=101, df_det="nom")
        )
        ## Select training set
        df_train = (
            df_data
            >> gr.tf_sample(n=10)
        )
        ## select test set
        df_test = (
            df_data
                >> gr.tf_anti_join(
                    df_train,
                    by=["x1", "x2"],
                )
                >> gr.tf_sample(n=200)
        )

        # Create fitted model
        md_fit = (
            df_train
            >> ft_gp(
                var=["x1", "x2", "x3"],
                out=["y1", "y2", "y3"],
            )
        )

        # Call eval_pnd
        df_pnd = (
            md_fit
            >> gr.ev_pnd(
                df_train,
                df_test,
                signs = {"y1":1, "y2":1,"y3":1},
                seed = 101
            )
        )

        # Test for correctness by shape
        self.assertTrue(len(df_pnd) == df_test.shape[0])
        # Test for correctness by # of outputs
        self.assertTrue(len(df_pnd.columns.values) == len(df_test.columns.values) + 2)

    def test_eval_append(self):
        """ Test append parameter on eval_pnd()
        """
        # Model to make Dataset
        md_true = make_pareto_random(twoDim=False)

        # Create dataframe
        df_data = (
            md_true
            >> gr.ev_sample(n=2e3, seed=101, df_det="nom")
        )
        ## Select training set
        df_train = (
            df_data
            >> gr.tf_sample(n=10)
        )
        ## select test set
        df_test = (
            df_data
                >> gr.tf_anti_join(
                    df_train,
                    by=["x1", "x2"],
                )
                >> gr.tf_sample(n=200)
        )

        # Create fitted model
        md_fit = (
            df_train
            >> ft_gp(
                var=["x1", "x2", "x3"],
                out=["y1", "y2", "y3"],
            )
        )

        # Call eval_pnd
        df_pnd = (
            md_fit
            >> gr.ev_pnd(
                df_train,
                df_test,
                signs = {"y1":1, "y2":1,"y3":1},
                seed = 101,
                append = False
            )
        )

        # Test for correctness by shape
        self.assertTrue(len(df_pnd) == df_test.shape[0])
        # Test for correctness by # of outputs
        self.assertTrue(len(df_pnd.columns.values) == 2)


    def test_eval_input_subsets(self):
        """ Test inputs are subsets of the provided DataFrames for eval_pnd()
        """
        # Model to make Dataset
        md_true = make_pareto_random(twoDim=False)

        # Create dataframe
        df_data = (
            md_true
            >> gr.ev_sample(n=2e3, seed=101, df_det="nom")
        )
        ## Select training set
        df_train = (
            df_data
            >> gr.tf_sample(n=10)
        )
        ## select test set
        df_test = (
            df_data
                >> gr.tf_anti_join(
                    df_train,
                    by=["x1", "x2"],
                )
                >> gr.tf_sample(n=200)
        )

        # Create fitted model
        md_fit = (
            df_train
            >> ft_gp(
                var=["x1", "x2", "x3"],
                out=["y1", "y2", "y3"],
            )
        )

        # Call eval_pnd w/ only "y1" and "y2"
        df_pnd = (
            md_fit
            >> gr.ev_pnd(
                df_train,
                df_test,
                signs = {"y1":1, "y2":1},
                seed = 101
            )
        )

        ### how to imply x1 and x2 from y1 and y2?

        # Test for correctness by shape
        self.assertTrue(len(df_pnd) == df_test.shape[0])
        # Test for correctness by # of outputs
        self.assertTrue(len(df_pnd.columns.values) == len(df_test.columns.values) + 2)

    def test_eval_faulty_inputs(self):
        """ Test faulty inputs to eval_pnd
        """
        # Model to make Dataset
        md_true = make_pareto_random()
        # Create dataframe
        df_data = (
            md_true
            >> gr.ev_sample(n=2e3, seed=101, df_det="nom")
        )
        ## Select training set
        df_train = (
            df_data
            >> gr.tf_sample(n=10)
        )
        ## select test set
        df_test = (
            df_data
                >> gr.tf_anti_join(
                    df_train,
                    by=["x1", "x2"],
                )
                >> gr.tf_sample(n=200)
        )

        # Create fitted model
        md_fit = (
            df_train
            >> ft_gp(
                var=["x1", "x2"],
                out=["y1", "y2"],
            )
        )

        # Call eval_pnd
        with self.assertRaises(ValueError):
            df_pnd = (
                md_fit
                >> gr.ev_pnd(
                    df_train,
                    df_test,
                    signs = {"y":1, "y2":1},
                    seed = 101
                )
            )


class TestUtilities(unittest.TestCase):

    def setUp(self):
        self.X_base = np.random.multivariate_normal(
            mean=[0, 0],
            cov=np.eye(2),
            size=50,
        )
        # Compute the pareto points
        self.idx_pareto = gr.pareto_min_rel(self.X_base)

    def test_pareto_min_rel(self):
        # Relative mode
        X_base = np.array([
            [1, 0],
            [0, 1],
        ])
        X_test = np.array([
            [0, 0],
            [1, 1],
        ])

        b_true = np.array([1, 0], dtype=bool)

        # Check accuracy
        self.assertTrue(
            np.all(gr.pareto_min_rel(X_test, X_base=X_base) == b_true)
        )

        # For Pareto frontier
        X_single = np.array([
            [2, 0],
            [1, 1],
            [0, 2],
            [2, 1],
            [2, 2],
            [1, 2],
        ])
        b_single = np.array([1, 1, 1, 0, 0, 0], dtype=bool)
        # Check accuracy
        res = gr.pareto_min_rel(X_single)
        self.assertTrue(
            np.all(res == b_single)
        )

    def test_fundamentals(self):
        # Test bare functionality
        Sigma = gr.make_proposal_sigma(self.X_base, self.idx_pareto, np.eye(2))
        X_sample = gr.rprop(10, Sigma, self.X_base[self.idx_pareto, :])
        d_sample = gr.dprop(X_sample, Sigma, self.X_base[self.idx_pareto, :])

    def test_pnd_order(self):
        """Test the PND MIS approximation for simple rank-ordering
        """
        # Against the common base-points, the following should be trivial to distinguish
        X_pred = np.array([
            [-1, -1],
            [ 0,  0],
            [+1, +1],
        ])
        X_cov = [np.eye(2)] * 3
        # A very coarse sample will suffice
        pr_scores, var_values = gr.approx_pnd(X_pred, X_cov, self.X_base, signs=np.array([-1, -1]), n=int(1e2))
        # Check basic ordering correctness
        self.assertTrue(pr_scores[0] > pr_scores[1])
        self.assertTrue(pr_scores[1] > pr_scores[2])

    def test_pnd_orthant(self):
        """Test the PND MIS approximation for accuracy with orthant tests
        """
        ## A single isotonic gaussian at the origin has a 75% chance of landing
        ## outside the negative orthant; we can use this as a numerical check of
        ## the PND MIS approximation
        X_pred = np.array([[0, 0]])
        X_cov = [np.eye(2) * 1e-1]
        X_base = np.array([
            [   -3,     0],
            [   -2,     0],
            [   -1,     0],
            [ -0.5,     0],
            [ -0.1,     0],
            [    0,     0],
            [    0,  -0.1],
            [    0,    -1],
            [    0,    -2],
            [    0,    -3],
        ])

        n = int(1e4)
        pr_scores, var_values = gr.approx_pnd(
            X_pred,
            X_cov,
            X_base,
            signs=np.array([+1, +1]),
            n=n,
            seed=101,
        )

        # Check for fixed level of accuracy
        self.assertTrue(np.abs(0.75 - pr_scores[0]) < 0.05)

        # Check that observed CI is compatible with true value
        pr_lo = pr_scores[0] - 2.58 * np.sqrt(var_values[0] / n) # 99% CI
        pr_hi = pr_scores[0] + 2.58 * np.sqrt(var_values[0] / n)
        self.assertTrue((pr_lo <= 0.75) and (0.75 <= pr_hi))

## Run tests
if __name__ == "__main__":
    unittest.main()


__all__ = [
    "tran_select",
    "tf_select",
    "tran_select_if",
    "tf_select_if",
    "tran_drop",
    "tf_drop",
    "tran_drop_if",
    "tf_drop_if",
    "is_numeric",
    "starts_with",
    "ends_with",
    "contains",
    "matches",
    "everything",
    "num_range",
    "one_of",
    "columns_between",
    "columns_from",
    "columns_to",
    "resolve_selection"
]

import re
from .base import Intention, dfdelegate, symbolic_evaluation, \
    group_delegation, flatten, make_symbolic
from .. import add_pipe
from numpy import zeros, where, ones
from numpy import max as npmax
from pandas import Index, Series
from pandas.api.types import is_numeric_dtype


# ------------------------------------------------------------------------------
# Select and drop operators
# ------------------------------------------------------------------------------


def selection_context(arg, context):
    if isinstance(arg, Intention):
        arg = arg.evaluate(context)
        if isinstance(arg, Index):
            arg = list(arg)
        if isinstance(arg, Series):
            arg = arg.name
    return arg


def selection_filter(f):
    def wrapper(*args, **kwargs):
        return Intention(
            lambda x: f(
                list(x.columns),
                *(selection_context(a, x) for a in args),
                **{k: selection_context(v, x) for k, v in kwargs.items()}
            )
        )

    return wrapper


def resolve_selection(df, *args, drop=False):
    if len(args) > 0:
        args = [a for a in flatten(args)]
        ordering = []
        column_indices = zeros(df.shape[1])
        for selector in args:
            visible = where(selector != 0)[0]
            if not drop:
                column_indices[visible] = selector[visible]
            else:
                column_indices[visible] = selector[visible] * -1
            for selection in where(selector == 1)[0]:
                if not df.columns[selection] in ordering:
                    ordering.append(df.columns[selection])
    else:
        ordering = list(df.columns)
        column_indices = ones(df.shape[1])
    return ordering, column_indices



@group_delegation
@symbolic_evaluation(eval_as_selector=True)
def tran_select(df, *args):
    r"""Select columns in a DataFrame

    Down-select or re-arrange columns in a DataFrame, usually for readability. Provide specific column names, or make use of the following selection helpers:

        starts_with() - column name begins with string
        ends_with() - column name ends with string
        contains() - column name contains string
        matches() - column name matches pattern (regular expression)
        everything() - all columns not already selected; useful for re-arranging columns without dropping

    Arguments:
        df (pandas.DataFrame): DataFrame to modify
        *args (str or selection helper): Specific column name OR selection helper.

    Returns:
        DataFrame: Data with selected columns

    Examples:

        ## Setup
        import grama as gr
        DF = gr.Intention()
        ## Load example dataset
        from grama.data import df_stang_wide

        ## Move "alloy" column to left
        (
            df_stang_wide
            >> gr.tf_select("alloy", gr.everything())
        )

        ## Find columns that start with "mu_"
        (
            df_stang_wide
            >> gr.tf_select(gr.starts_with("mu"))
        )

        ## Find columns with digits in names
        (
            df_stang_wide
            >> gr.tf_select(gr.matches("\\d+"))
        )

    """
    ordering, column_indices = resolve_selection(df, *args)
    if (column_indices == 0).all():
        return df[[]]
    selection = where(
        (column_indices == npmax(column_indices)) & (column_indices >= 0)
    )[0]
    df = df.iloc[:, selection]
    if all([col in ordering for col in df.columns]):
        ordering = [c for c in ordering if c in df.columns]
        return df[ordering]
    return df

tf_select = add_pipe(tran_select)


@group_delegation
@symbolic_evaluation(eval_as_selector=True)
def tran_drop(df, *args):
    _, column_indices = resolve_selection(df, *args, drop=True)
    if (column_indices == 0).all():
        return df[[]]
    selection = where(
        (column_indices == npmax(column_indices)) & (column_indices >= 0)
    )[0]
    return df.iloc[:, selection]

tf_drop = add_pipe(tran_drop)


@dfdelegate
def tran_select_if(df, fun):
    """Selects columns where fun(ction) is true
    Args:
        fun: a function that will be applied to columns
    """

    def _filter_f(col):
        try:
            return fun(df[col])
        except:
            return False

    cols = list(filter(_filter_f, df.columns))
    return df[cols]

tf_select_if = add_pipe(tran_select_if)


@dfdelegate
def tran_drop_if(df, fun):
    """Drops columns where fun(ction) is true
    Args:
        fun: a function that will be applied to columns
    """

    def _filter_f(col):
        try:
            return fun(df[col])
        except:
            return False

    cols = list(filter(_filter_f, df.columns))
    return df.drop(cols, axis=1)

tf_drop_if = add_pipe(tran_drop_if)

@selection_filter
def starts_with(columns, prefix):
    r"""Select columns starting with a prefix, for use in tran_select()

    Args:
        prefix (str): Prefix to detect

    """
    return [c for c in columns if c.startswith(prefix)]


@selection_filter
def ends_with(columns, suffix):
    r"""Select columns ending in a suffix, for use in tran_select()

    Args:
        suffix (str): Suffix to detect

    """
    return [c for c in columns if c.endswith(suffix)]


@selection_filter
def contains(columns, substr):
    r"""Select columns containing a substring, for use in tran_select()

    Args:
        substr (str): Substring to detect

    """
    return [c for c in columns if substr in c]


@selection_filter
def matches(columns, pattern):
    r"""Select columns matching a pattern, for use in tran_select()

    Args:
        pattern (str): String pattern to match, can be a regular expression

    """
    return [c for c in columns if re.search(pattern, c)]


@selection_filter
def everything(columns):
    "Select all columns, for use in tran_select()"
    return columns


@selection_filter
def num_range(columns, prefix, range):
    colnames = [prefix + str(i) for i in range]
    return [c for c in columns if c in colnames]


@selection_filter
def one_of(columns, specified):
    return [c for c in columns if c in specified]


@selection_filter
def columns_between(columns, start_col, end_col, inclusive=True):
    if isinstance(start_col, str):
        start_col = columns.index(start_col)
    if isinstance(end_col, str):
        end_col = columns.index(end_col)
    return columns[start_col : end_col + int(inclusive)]


@selection_filter
def columns_from(columns, start_col):
    if isinstance(start_col, str):
        start_col = columns.index(start_col)
    return columns[start_col:]


@selection_filter
def columns_to(columns, end_col, inclusive=False):
    if isinstance(end_col, str):
        end_col = columns.index(end_col)
    return columns[: end_col + int(inclusive)]


@make_symbolic
def is_numeric(series):
    """Determine if column is numeric

    Returns True if the provided column is numeric, False otherwise. Intended for calls to select_if() or mutate_if().

    Args:
        bool: Boolean corresponding to the datatype of the given column

    Examples::
        import grama as gr
        from grama.data import df_diamonds

        (
            df_diamonds
            gr.tf_select_if(gr.is_numeric)
        )

    """
    return is_numeric_dtype(series)



__all__ = [
    "tran_summarize",
    "tf_summarize",
    "tran_summarize_each",
    "tf_summarize_each",
]

from .base import dfdelegate
from .. import add_pipe
from pandas import DataFrame, Series


@dfdelegate
def tran_summarize(df, **kwargs):
    r"""Compute specified summaries

    Compute summaries: Takes values across multiple rows and combines them into fewer rows using the functions that you specify. If the DataFrame is not grouped, then there will be just one resulting row. If the DataFrame is grouped, then there will be one row per group.

    Use tran_group_by() to group by column values before a summarize.

    Use the Intention operator (usually `DF = gr.Intention()`) as a convenient way to access columns in the DataFrame.

    Some useful summary functions are:
    - mean(x) - compute the mean, see mean_lo() and mean_up() for confidence interval bounds
    - median(x) - compute the median
    - sd(x) - compute the standard deviation
    - IQR(x) - compute the interquartile range (IQR)
    - pr(x) - compute a probability, see pr_lo() and pr_up() for confidence interval bounds
    - corr(x, y) - compute the correlation between x and y

    Args:
        df (pandas.DataFrame): Data to summarize
        **kwargs (array-like or summary expression): Summaries to assign;
            the name of the argument (left of `=`) will be the new column name,
            the value of the argument (right of `=`) defines the summary

    Returns:
        DataFrame: Summarized data

    Examples::

        ## Setup
        import grama as gr
        DF = gr.Intention()
        ## Load example dataset
        from grama.data import df_diamonds

        ## Compute some summary statistics
        (
            df_diamonds
            >> gr.tf_summarize(
                carat_mean=gr.mean(DF.carat),
                carat_median=gr.median(DF.carat),
                carat_sd=gr.sd(DF.carat),
                carat_iqr=gr.IQR(DF.carat),
            )
        )

        ## Compute statistics across cuts
        (
            df_diamonds
            >> gr.tf_group_by(DF.cut)
            >> gr.tf_summarize(
                carat_mean=gr.mean(DF.carat),
                carat_median=gr.median(DF.carat),
                carat_sd=gr.sd(DF.carat),
                carat_iqr=gr.IQR(DF.carat),
            )
        )

    """
    return DataFrame({k: [v] for k, v in kwargs.items()})

tf_summarize = add_pipe(tran_summarize)


@dfdelegate
def tran_summarize_each(df, functions, *args):
    columns, values = [], []
    for arg in args:
        if isinstance(arg, Series):
            varname = arg.name
            col = arg
        elif isinstance(arg, str):
            varname = arg
            col = df[varname]
        elif isinstance(arg, int):
            varname = df.columns[arg]
            col = df.iloc[:, arg]

        for f in functions:
            fname = f.__name__
            columns.append("_".join([varname, fname]))
            values.append(f(col))

    return DataFrame([values], columns=columns)

tf_summarize_each = add_pipe(tran_summarize_each)



import numpy as np
import pandas as pd
import unittest
import io
import sys

from context import grama as gr
from context import models

## Test the built-in models
##################################################
class TestModels(unittest.TestCase):

    def setUp(self):
        pass

    def test_make(self):
        ## Models build
        md_cantilever_beam = models.make_cantilever_beam()
        md_ishigami = models.make_ishigami()
        md_linear_normal = models.make_linear_normal()
        md_piston = models.make_piston()
        md_piston_rand = models.make_piston_rand()
        md_plane_laminate = models.make_composite_plate_tension([0])
        md_plate_buckling = models.make_plate_buckle()
        md_poly = models.make_poly()
        md_prlc = models.make_prlc()
        md_prlc_rand = models.make_prlc_rand()
        md_test = models.make_test()
        md_trajectory_linear = models.make_trajectory_linear()

        ## Models evaluate
        df_cantilever = md_cantilever_beam >> gr.ev_nominal(df_det="nom")
        df_ishigami = md_ishigami >> gr.ev_nominal(df_det="nom")
        df_ln = md_linear_normal >> gr.ev_nominal(df_det="nom")
        df_piston = md_piston >> gr.ev_nominal(df_det="nom")
        df_piston_rand = md_piston_rand >> gr.ev_nominal(df_det="nom")
        df_plane = md_plane_laminate >> gr.ev_nominal(df_det="nom")
        df_plate = md_plate_buckling >> gr.ev_nominal(df_det="nom")
        df_poly = md_poly >> gr.ev_nominal(df_det="nom")
        df_prlc = md_prlc >> gr.ev_nominal(df_det="nom")
        df_prlc_rand = md_prlc_rand >> gr.ev_nominal(df_det="nom")
        df_test = md_test >> gr.ev_nominal(df_det="nom")
        df_traj = md_trajectory_linear >> gr.ev_nominal(df_det="nom")

        ## Piston models give approximately same output
        self.assertTrue(abs(df_piston.t_cyc[0] - df_piston_rand.t_cyc[0]) < 1e-6)

    def test_sir(self):
        from numpy import real
        from scipy.special import lambertw
        ## Verification test
        # Test parameters
        I0 = 1
        S0 = 99
        R0 = 0
        beta = 0.5
        gamma = 0.2

        # Asymptotic solution parameters
        N = I0 + S0 + R0
        R_0 = beta / gamma
        s_0 = S0 / N
        r_0 = R0 / N

        # Asymptotic solution
        S_inf = real(-(1/R_0) * lambertw(-s_0 * R_0 * np.exp(-R_0 * (1 - r_0))) * N)

        ## Base tolerance
        md_sir = models.make_sir()
        df_inf = gr.eval_df(
            md_sir,
            gr.df_make(
                t=1e6, # Approximation of t -> +\infty
                I0=I0,
                N=N,
                beta=beta,
                gamma=gamma,
            )
        )
        S_inf_comp = df_inf.S.values[-1]

        # Check relative tolerance
        self.assertTrue(abs(S_inf - S_inf_comp) / S_inf < 1e-3)
        self.assertTrue(abs(S_inf - S_inf_comp) / S_inf > 1e-5)

        ## Refined tolerance
        md_sir = models.make_sir(rtol=1e-6)
        df_inf = gr.eval_df(
            md_sir,
            gr.df_make(
                t=1e6, # Approximation of t -> +\infty
                I0=I0,
                N=N,
                beta=beta,
                gamma=gamma,
            )
        )
        S_inf_comp = df_inf.S.values[-1]

        # Check relative tolerance
        self.assertTrue(abs(S_inf - S_inf_comp) / S_inf < 1e-5)
